from subprocess import Popen, PIPE
import subprocess
import re, os
import numpy as np
import pandas as pd
import geoip2.database
from itertools import groupby
import ipaddress
import multiprocessing
from multiprocessing import Process, Manager
#import cymruwhois
#from cymruwhois import Client
import json 
import csv 
import ipaddress

# pd.options.mode.chained_assignment = None
maxmind_reader = geoip2.database.Reader('./input/maxmind/GeoLite2-City.mmdb')
cc_reader = geoip2.database.Reader('./input/maxmind/GeoLite2-Country.mmdb')

# ---- MULTIPROCESSING FUNCTIONS ----------

def run_command(command):
    return subprocess.Popen(command,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.STDOUT,
                         shell = True)

def _apply_df(args):
    df, func, kwargs = args
    return df.apply(func, **kwargs)

def apply_by_multiprocessing(df, func, **kwargs):
    workers = kwargs.pop('workers')
    pool = multiprocessing.Pool(processes=workers)
    result = pool.map(_apply_df, [(d, func, kwargs)
            for d in np.array_split(df, workers)])
    pool.close()
    return pd.concat(list(result))

# def apply_by_multiprocessing(df,func,**kwargs):
#   workers=kwargs.pop('workers')
#   pool = multiprocessing.Pool(processes=workers)
#   result = pool.map(_apply_df, [(d, func, i, kwargs) for i,d in enumerate(np.array_split(df, workers))])  
#   pool.close()
#   result=sorted(result,key=lambda x:x[0])
#   return pd.concat([i[1] for i in result])

def apply_by_multiproc_ordered(df,func,**kwargs):
    workers=kwargs.pop('workers')
    pool = multiprocessing.Pool(processes=workers)
    result = pool.map(_apply_df, [(d, func, i, kwargs) for i,d in enumerate(np.array_split(df, workers))])  
    pool.close()
    result=sorted(result,key=lambda x:x[0])
    return pd.concat([i[1] for i in result])
    return iter(p.stdout.readline, b'')

# ------------------GEOLOCATION FUNCTIONS-----------------------

def get_CC(row):
    if ipaddress.ip_address(unicode(str(row.A_record))).is_private:
        return None
    else:
        try:
            geoinfo = cc_reader.country(str(row.A_record))
            return geoinfo.country.iso_code
        except:
            return None

def get_city(row):
    if ipaddress.ip_address(unicode(str(row.A_record))).is_private:
        return None
    else:
        try:
            geoinfo = maxmind_reader.city(str(row.A_record))
            return geoinfo.city.name
        except:
            return None

def get_location(row):
    #TODO: add here the dig query -- get the PTR, parse the hostname for airport codes, etc.
    if ipaddress.ip_address(unicode(str(row.A_record))).is_private:
        return None
    else:
        try:
            geoinfo = maxmind_reader.city(str(row.A_record))
            return (geoinfo.location.latitude, geoinfo.location.longitude)
        except:
            return None


# --------------------------------------------------------------

# use as test input the file at dns_detect_parser/test_input/IT_TIM/4G/427/1267roam/285_MONROE.EXP.DIG_2_1513753795.json
'''
{
  "couponsvc.com": {
    "default": {
      "publicIPSubnet": "90.174.3.54",
      "when_human": "Wed Dec 20 07:10:04 UTC 2017",
      "msgsize": "70",
      "Timestamp": 1513753803,
      "queryTime": "90",
      "server": "85.62.229.135#53(85.62.229.135)",
      "raw": "\n; <<>> DiG 9.10.3-P4-Debian <<>> -b 172.18.21.2 couponsvc.com +subnet=90.174.3.54\/32\n
      			;; global options: +cmd\n;; Got answer:\n;; ->>HEADER<<- opcode: QUERY, status: NOERROR, id: 32734\n
      			;; flags: qr rd ra; QUERY: 1, ANSWER: 1, AUTHORITY: 0, ADDITIONAL: 1\n\n
      			;; OPT PSEUDOSECTION:\n
      			; EDNS: version: 0, flags:; udp: 512\n
      			; CLIENT-SUBNET: 90.174.3.54\/32\/0\n
      			;; QUESTION SECTION:\n
      			;couponsvc.com.\t\t\tIN\tA\n\n
      			;; ANSWER SECTION:\n
      			couponsvc.com.\t\t300\tIN\tA\t69.172.201.153\n\n
      			;; Query time: 90 msec\n;; SERVER: 85.62.229.135#53(85.62.229.135)\n
      			;; WHEN: Wed Dec 20 07:10:04 UTC 2017\n
      			;; MSG SIZE  rcvd: 70\n\n",
      "digClientSubnetFlag": false,
      "digversion": "9.10.3",
      "answer": {
        "question": [
          
        ],
        "additional": [
          
        ],
        "authority": [
          
        ],
        "ans": [
          {
            "rightSide": "69.172.201.153",
            "typeOfRecord": "A",
            "leftSide": "couponsvc.com.",
            "TTL": "300"
          }
        ]
      }
    }
  }
}
'''
# create pands data frame to hold the information from the DNS_DETECT tests
cols = ['Iccid', 'Operator', 'NodeId', 'IMSIMCCMNC', 'NWMCCMNC', 'RAT', 'publicIP', 'ABP_target', 'msgsize', 'Timestamp', 'queryTime', 'server', 'A_record', 'TTL', 'fqdn']

# read the list of target ABP FQDNs the we resolve in this test
ABP_FQDN = []
ABP_file = open ("./test_input/ABP_filters_fqdn_list", 'r')
for line in ABP_file:
	ABP_FQDN.append(line.strip())

IN_DIR = "/storage/ROAM/ROAM_MNO_RAT/dns_detect/"
#IN_DIR = "/Users/andra/Documents/MY_WORK/ROAMING/GIT/roaming/parsers/dns_detect_parser/test_input/IT_TIM/4G/"
# TODO: change this to traverse the whole set of results
#test_input = "/Users/andra/Documents/MY_WORK/ROAMING/GIT/roaming/parsers/dns_detect_parser/test_input/IT_TIM/4G/1267roam/285_MONROE.EXP.DIG_2_1513753795.json"

# each .json file contains the results of one single run of the dns_detect container
# define function to read the content from each json file
# the function returns a small data frame, which we store in a list 

def read_json(in_results, RAT):
    dns_frame = pd.DataFrame([], columns = cols)
    print str(in_results)
    dns_data = {}
    with open(in_results) as json_data:
        dns_data = json.load(json_data) 
    print ("----- MONROE Info ----- ")
    print("ICCID: " + str(dns_data['Iccid']))
    iccid = dns_data['Iccid']
    print ("Operator: " + str(dns_data['Operator']))
    operator = dns_data['Operator']
    print ("IMSIMCCMNC: " + str(dns_data['IMSIMCCMNC']))
    imsimccmnc = dns_data['IMSIMCCMNC']
    print ("NWMCCMNC: " + str(dns_data['NWMCCMNC']))
    nwmccmnc = dns_data['NWMCCMNC']
    print ("Public IP: " +  str(dns_data['publicIP']))
    publicIP = dns_data['publicIP'].strip("\n")
    nodeid = dns_data['NodeId']
    rat = RAT
    monroe = [iccid, operator, nodeid, imsimccmnc, nwmccmnc, rat, publicIP]
    print ("------------------------ ")
    #dns_detect = pd.DataFrame(np.array([r]), columns=cols)
    for target in dns_data:
      if target in ABP_FQDN:
        # print "DNS server used: " + str([str(x) for x in dns_data[target]])
        if 'default' in dns_data[target]:
          dig = dns_data[target]['default']
          # bad formatting for 'publicIPSubnet',
          rec = [target, dig['msgsize'], dig['Timestamp'], dig['queryTime'], dig['server'].split("#")[0]]
          if dig['answer']['ans']:
              # print target + ":"
              # print dig['answer']['ans']
              for A in dig['answer']['ans']:
                if A['typeOfRecord'] == "A":
                  A_record = A['rightSide']
                  ttl = A['TTL']
                  fqdn = A['leftSide']
                  r = monroe + rec + [A_record, ttl, fqdn]
                  #print str(r)
                  row = pd.DataFrame(np.array([r]), columns=cols)
                  dns_frame = pd.concat([dns_frame, row], ignore_index=True)
                  #nsb = pd.DataFrame(np.array([rec]), columns=cols).append(nsb, ignore_index=True)
          else:
            continue
            #print target + ": no DNS reply"
        else:
          continue
      else:
        continue
    # with open("dns_detect_test.csv", 'a') as f:
    #     dns_frame.to_csv(f, sep = ";", index=False, header=False)
    return dns_frame

frames = []

dns_detect = pd.DataFrame([], columns = cols) # empty dataframe that we should populate

for dirName, subdirList, fileList in os.walk(IN_DIR):
    print('Found directory: %s' % dirName)
    for fname in fileList:
      if ".json" in fname and "4G" in dirName:
            print "-----------------4G-----------------"
            print('Parsing the following results file\n\t%s' % fname)
            f = str(dirName) + "/" + str(fname)
            print ('Loading the results from ' + str(f))
            frames.append(read_json(f, "4G"))
      if ".json" in fname and "3G" in dirName:
            print "-----------------3G-----------------"
            print('Parsing the following results file\n\t%s' % fname)
            f = str(dirName) + "/" + str(fname)
            print ('Loading the results from ' + str(f))
            frames.append(read_json(f, "3G"))
            #dns_detect = pd.concat([dns_detect, read_json(f)], axis=0, ignore_index=True)

dns_detect = pd.concat(frames, ignore_index=True)
del(frames)

# sample format and content of the dns_detect frame
# Iccid;Operator;NodeId;IMSIMCCMNC;NWMCCMNC;RAT;publicIP;ABP_target;msgsize;Timestamp;queryTime;server;A_record;TTL;fqdn
# 89390100002191720550;I TIM;328;22201;24202;4G;62.18.220.21;theprivilegesbox.com;65;1513829100;223;213.230.134.106;45.33.9.234;300;theprivilegesbox.com.
# 89390100002191720550;I TIM;328;22201;24202;4G;62.18.220.21;iabfjs.info;56;1513829082;121;213.230.134.106;31.199.53.10;0;iabfjs.info.

dns_detect['city'] = apply_by_multiprocessing(dns_detect, get_city, axis=1, workers=60)
print "finished getting the server city"
dns_detect['CC'] = apply_by_multiprocessing(dns_detect, get_CC, axis=1, workers=60)
print "finished getting the CC"
dns_detect['GPS_loc'] = apply_by_multiprocessing(dns_detect, get_location, axis=1, workers=60)
print "finished getting the GPS location"

# save to disk the data frame with all the measurements we retrieved from the JSON output
dns_detect.to_csv("dns_detect_ALL.csv", sep = ";", index = False)







